from .models.module_models import ModuleUpdate as ModuleUpdate
from .models.windprobe_models import WindProbeData as WindProbeData
from .models.windtrack_models import TrackingData as TrackingData
from .models.windtrack_models import TrackingDataDict as TrackingDataDict
from .windsuite_sdk import WindsuiteSDK as WindsuiteSDK



"""
Represnts the full windsuite object,
This will contain the other user SDKs.


"""

from __future__ import annotations

import asyncio
import threading
import time
from collections.abc import Callable
from typing import Any, Protocol

import socketio  # type: ignore[missing stub]
from loguru import logger
from pydantic import ValidationError

from .fan_controller import FanController, Indexed80mmModule, Indexed240mmModule
from .models.module_models import ModuleUpdate
from .models.windprobe_models import WindProbeData
from .models.windtrack_models import TrackingDataDict


class CallbackWindtrack(Protocol):
    """
    This represents a callback function that the user will provide to handle windtrack data.

    It will be called each time a new windtrack data is received, given the data as argument.

    Args:
        data (TrackingDataDict): A dictionary of tracking data, with the object name as key and the TrackingData as value.

    """

    def __call__(self, data: TrackingDataDict) -> None: ...


class CallbackWindprobe(Protocol):
    """
    This represents a callback function that the user will provide to handle windprobe data.

    It will be called each time a new windprobe data is received, given the data as argument.

    Args:
        data (WindProbeData): The windprobe data received.

    """

    def __call__(self, data: WindProbeData) -> None: ...


class CallbackModuleUpdate(Protocol):
    """
    This represents a callback function that the user will provide to handle module update data.

    It will be called each time a new module update data is received, given the data as argument.

    Args:
        data (dict): The module update data received.

    """

    def __call__(self, module_updates: dict[str, ModuleUpdate]) -> None: ...


class WindsuiteSDK:
    def __init__(
        self,
        base_url: str,
        port_rest_api: int = 8000,
        port_socketio_api: int = 8001,
        machine_size: tuple[int, int] = (2, 2),
    ) -> None:
        self.port_rest_api = port_rest_api
        self.port_socketio_api = port_socketio_api
        self.rest_base_url = base_url.rstrip("/") + f":{self.port_rest_api}" + "/api"
        self.socketio_base_url = base_url.rstrip("/") + f":{self.port_socketio_api}"
        self.machine_size = machine_size

        self.__callback_windtrack: CallbackWindtrack | None = None
        self.__callback_windprobe: CallbackWindprobe | None = None
        self.__callback_module_update: CallbackModuleUpdate | None = None

        self.fan_controller = FanController(base_url=self.rest_base_url, machine_size=self.machine_size)

        self.stop_event = threading.Event()
        self.__running_threads: list[threading.Thread] = []

    def start_socketio_thread(self) -> None:
        """
        Starts the __socketio_client_routine in a background thread, handling event loop issues.
        """

        def run_socketio() -> None:
            try:
                # ! NOT SUPPORTED ON WINDOWS
                # asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
                loop: asyncio.AbstractEventLoop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop=loop)
                loop.run_until_complete(future=self.__socketio_client_routine())
            except RuntimeError as e:
                logger.info(f"SocketIO thread exception: {e}")

        t = threading.Thread(target=run_socketio, daemon=True)
        t.name = "socketio_client_thread"
        self.__running_threads.append(t)
        t.start()

    def stop_everything(self) -> None:
        """
        Stops the socketio client and sets the stop event.
        """
        self.stop_event.set()
        for thread in self.__running_threads:
            if thread.is_alive():
                thread.join()

    def create_80mm_module_to_set(
        self,
        index_xy: tuple[int, int],
        pwms_downstream_layer: list[float] | float,
        pwms_upstream_layer: list[float] | float,
    ) -> Indexed80mmModule:
        """
        If a single float is provided for a layer, it will be expanded to all fans in that layer.

        Otherwise, the len HAS to be fans_per_layer.
        """
        return self.fan_controller.create_80mm_module_to_set(
            index_xy=index_xy,
            pwms_downstream_layer=pwms_downstream_layer,
            pwms_upstream_layer=pwms_upstream_layer,
        )

    def create_240mm_module_to_set(
        self,
        index_xy: tuple[int, int],
        pwm: float,
    ) -> Indexed240mmModule:
        return self.fan_controller.create_240mm_module_to_set(
            index_xy=index_xy,
            pwm=pwm,
        )

    def set_fans(self, list_indexed_modules: list[Indexed80mmModule | Indexed240mmModule]) -> None:
        return self.fan_controller.set_fans(list_indexed_modules=list_indexed_modules)

    def set_psus(self, psu_state: bool) -> None:
        return self.fan_controller.set_psus(psu_state=psu_state)

    async def __socketio_client_routine(self) -> None:
        """
        Create a SocketIO client to connect to the Windsuite API.

        Contains the routes that receive from the different events

        This is started in a thread and will run until the stop_event is set
        """
        sio = socketio.AsyncClient()

        @sio.event  # type: ignore[UnknownMemberType], Rationale : type hints of socketio are not strict enough
        async def connect() -> None:  # type: ignore[reportUnusedFunction], Rationale: Event callback, not called directly
            logger.info("Connected to SocketIO server.")

        @sio.event  # type: ignore[UnknownMemberType], Rationale : type hints of socketio are not strict enough
        async def disconnect() -> None:  # type: ignore[reportUnusedFunction], Rationale: Event callback, not called directly
            logger.info("Disconnected from SocketIO server.")

        @sio.event  # type: ignore[UnknownMemberType], Rationale : type hints of socketio are not strict enough
        async def probe_data(data: str) -> None:  # type: ignore[reportUnusedFunction], Rationale: Event callback, not called directly
            # logger.info(f"Received message: {data}")
            try:
                windprobe_data = WindProbeData.model_validate(data)
                if self.__callback_windprobe:
                    self.__callback_windprobe(windprobe_data)
            except ValidationError as e:
                logger.error(f"Validation error: {e}")

        @sio.event  # type: ignore[UnknownMemberType], Rationale : type hints of socketio are not strict enough
        async def tracking_data(data: dict) -> None:  # type: ignore[reportUnusedFunction], Rationale: Event callback, not called directly
            # print(f"Received windtrack data: {data}")
            try:
                windtrack_dict = TrackingDataDict.model_validate(data)
                if self.__callback_windtrack:
                    self.__callback_windtrack(windtrack_dict)
            except ValidationError as e:
                print(f"Validation error: {e}")

        @sio.event  # type: ignore[UnknownMemberType], Rationale : type hints of socketio are not strict enough
        async def module_update(data: dict[Any]) -> None:  # type: ignore[reportUnusedFunction], Rationale: Event callback, not called directly
            try:
                all_data = data["updated_modules"]
                # logger.debug(f"Received module update data: {all_data}")
                validated_data: dict[str, ModuleUpdate] = {}
                for mac, module_dict in all_data.items():
                    module_update = ModuleUpdate(**module_dict)
                    validated_data[mac] = module_update

                if self.__callback_module_update:
                    self.__callback_module_update(module_updates=validated_data)

            except ValidationError as e:
                print(f"Validation error: {e}")

        try:
            await sio.connect(  # type: ignore[UnknownMemberType], Rationale : type hints of socketio are not strict enough
                self.socketio_base_url,
                transports=["websocket"],
            )
        except socketio.exceptions.ConnectionError as e:  # type: ignore[UnknownMemberType], Rationale : type hints of socketio are not strict enough
            print(
                f"Socket IO connection error: {e},\nHINT :\n\t- Verify that the server is running\n\t- The URL is correct and\n\t- You're on the same network as the server"
            )
            self.stop_event.set()
            return

        # Wait for stop_event in a non-blocking way and disconnect sio when set
        while not self.stop_event.is_set():  # noqa: ASYNC110, Rationale: honestly, seems to be the best way to do this
            await asyncio.sleep(0.1)
        await sio.disconnect()

    def register_tracking_callback(
        self,
        callback: CallbackWindtrack,
    ) -> None:
        """
        Register a callback for windtrack data.

        Args:
            callback (Callable[[TrackingData], None]): The function to call when windtrack data is received.

        """
        if not callable(callback):
            raise TypeError("Callback must be a callable function.")

        self.__callback_windtrack = callback

    def register_windprobe_callback(
        self,
        callback: CallbackWindprobe,
    ) -> None:
        """
        Register a callback for windprobe data.

        Args:
            callback (CallbackWindprobe): The function to call when windprobe data is received.

        """
        if not callable(callback):
            raise TypeError("Callback must be a callable function.")

        self.__callback_windprobe = callback

    def register_module_update_callback(
        self,
        callback: CallbackModuleUpdate,
    ) -> None:
        """
        Register a callback for module update data.

        Args:
            callback (CallbackModuleUpdate): The function to call when module update data is received.

        """
        if not callable(callback):
            raise TypeError("Callback must be a callable function.")

        self.__callback_module_update = callback

    def main_loop_manager(
        self,
        main_function: Callable[..., None],
        loop_frequency_hz: float,
    ) -> None:
        """
        Utility function that will call the given function at the given frequency until stop flag is set.

        This is given to users of the SDK, so they don't have to implement their own loop.

        This has the advantage of being cleanly stoppable with Ctrl+C and will stop all the SDK background stuff properly

        ! Don't use delays or blocking calls in the main_function, as this will block the loop.

        Args:
            main_function (callable): The function to run in the loop.
            loop_frequency_hz (float): Frequency in Hz to run the main_function.

        """
        try:
            period = 1.0 / loop_frequency_hz
            remaining = period

            while not self.stop_event.wait(timeout=remaining):
                start_ns = time.monotonic_ns()

                main_function()

                elapsed_ns = time.monotonic_ns() - start_ns
                remaining = max(0, period - (elapsed_ns / 1e9))

        except KeyboardInterrupt:
            print("\nExiting...")
        finally:
            self.stop_everything()


if __name__ == "__main__":
    pass



from collections.abc import Iterator

from pydantic import BaseModel


class Vec3(BaseModel):
    x: float
    y: float
    z: float

    def __iter__(self) -> Iterator[float]:  # type: ignore[IncompatibleMethodOverride]
        yield self.x
        yield self.y
        yield self.z

    def __getitem__(self, index: int) -> float:
        if index == 0:
            return self.x
        if index == 1:
            return self.y
        if index == 2:
            return self.z
        raise IndexError("Index out of range for Vec3, valid indices are 0, 1, 2.")


class Quat(BaseModel):
    w: float
    x: float
    y: float
    z: float

    def __iter__(self) -> Iterator[float]:  # type: ignore[IncompatibleMethodOverride]
        yield self.w
        yield self.x
        yield self.y
        yield self.z

    def __getitem__(self, index: int) -> float:
        if index == 0:
            return self.w
        if index == 1:
            return self.x
        if index == 2:
            return self.y
        if index == 3:
            return self.z
        raise IndexError("Index out of range for Quat, valid indices are 0, 1, 2, 3.")



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto

import requests
from loguru import logger


class FanType(Enum):
    FAN_80MM = auto()
    FAN_240MM = auto()


@dataclass
class Indexed80mmModule:
    flat_index: int
    downstream_layer: list[int]
    upstream_layer: list[int]


@dataclass
class Indexed240mmModule:
    flat_index: int
    pwm: int


class FanController:
    def __init__(self, base_url: str, machine_size: tuple[int, int]) -> None:
        self.base_url = base_url.rstrip("/")
        self.machine_size = machine_size

    def create_80mm_module_to_set(
        self,
        index_xy: tuple[int, int],
        pwms_downstream_layer: list[float] | float,
        pwms_upstream_layer: list[float] | float,
    ) -> Indexed80mmModule:
        if index_xy[0] >= self.machine_size[0] or index_xy[1] >= self.machine_size[1]:
            logger.error(f"Index {index_xy} out of range of the machine size {self.machine_size}")
            raise ValueError("Index out of range of the machine")

        flat_index = index_xy[0] * self.machine_size[1] + index_xy[1]

        if isinstance(pwms_downstream_layer, (float, int)):
            pwms_downstream_layer = [pwms_downstream_layer] * 9
        if isinstance(pwms_upstream_layer, (float, int)):
            pwms_upstream_layer = [pwms_upstream_layer] * 9

        pwms_downstream_layer_int = [round(max(0.0, min(100.0, pwm)) * 10) for pwm in pwms_downstream_layer]
        pwms_upstream_layer_int = [round(max(0.0, min(100.0, pwm)) * 10) for pwm in pwms_upstream_layer]

        return Indexed80mmModule(
            flat_index=flat_index,
            downstream_layer=pwms_downstream_layer_int,
            upstream_layer=pwms_upstream_layer_int,
        )

    def create_240mm_module_to_set(
        self,
        index_xy: tuple[int, int],
        pwm: float,
    ) -> Indexed240mmModule:
        if index_xy[0] >= self.machine_size[0] or index_xy[1] >= self.machine_size[1]:
            logger.error(f"Index {index_xy} out of range of the machine size {self.machine_size}")
            raise ValueError("Index out of range of the machine")

        if not isinstance(pwm, float):
            logger.error("For 240mm fans, pwm has to be a single float value")
            raise TypeError("For 240mm fans, pwm has to be a single float value")

        flat_index = index_xy[0] * self.machine_size[1] + index_xy[1]

        pwm_int = round(max(0.0, min(100.0, pwm)) * 10)

        return Indexed240mmModule(
            flat_index=flat_index,
            pwm=pwm_int,
        )

    def set_fans(self, list_indexed_modules: list[Indexed80mmModule | Indexed240mmModule]) -> None:
        """
        This will take a list of IndexedModule to set
        and call the REST api to set the fans accordingly
        """
        pwm_data: dict[str, dict[str, list[int]]] = {}

        for module in list_indexed_modules:
            pwm_data[str(module.flat_index)] = {}
            if isinstance(module, Indexed240mmModule):
                pwm_data[str(module.flat_index)]["0"] = [module.pwm]

            elif isinstance(module, Indexed80mmModule):  # type: ignore[unreachable], Rationale: could be reachable if type hints are ignored
                if module.downstream_layer:
                    pwm_data[str(module.flat_index)]["0"] = module.downstream_layer
                if module.upstream_layer:
                    pwm_data[str(module.flat_index)]["1"] = module.upstream_layer
            else:
                logger.error(f"Unknown module type: {type(module)}")
                raise TypeError("Unknown module type")

        payload = {"pwm_data": pwm_data}

        url = f"{self.base_url}/windcontrol/modules/pwms"
        # logger.error(f"Setting fans with payload: {payload} at {url}")
        response = requests.patch(url, json=payload, timeout=5)
        response.raise_for_status()
        if response.status_code == 200:
            pass
            # logger.info("Successfully set fans")
        else:
            logger.error(f"Set fans response: {response.status_code} - {response.text}")

    def set_psus(self, psu_state: bool):
        """
        Sets the PSU state via a POST request.
        """
        url = f"{self.base_url}/windcontrol/modules/psu"
        params = {"psu_state_to_set": str(psu_state).lower()}
        logger.debug(f"PATCH {url} with params={params}")
        response = requests.patch(url, params=params, timeout=5)
        response.raise_for_status()
        if response.status_code == 200:
            pass
        else:
            logger.error(f"Set PSU response: {response.status_code} - {response.text}")



from pydantic import BaseModel

from .common_models import Quat, Vec3

tracked_object_name = str


class TrackingData(BaseModel):
    """
    Represents tracking data for a single object.

    - Position in meters in the world reference frame,
    - Orientation as a quaternion in the world reference frame,

    ! Special case for the windshaper itself that will have the same as it's his reference frame
    - Position in meters in the windshaper reference frame,
    - Orientation as a quaternion in the windshaper reference frame,

    - Velocity in meters per second,
    - Angular velocity in degrees per second.
    """

    timestamp: float

    position_meters_world_ref: Vec3
    rotation_world_ref: Quat

    position_meters_windshaper_ref: Vec3
    rotation_windshaper_ref: Quat

    velocity_mps_world_ref: Vec3
    angular_velocity_degps_world_ref: Vec3

    velocity_mps_windshaper_ref: Vec3
    angular_velocity_degps_windshaper_ref: Vec3


class TrackingDataDict(BaseModel):
    data: dict[tracked_object_name, TrackingData]



from __future__ import annotations

from collections.abc import Sequence
from enum import Enum, auto
from typing import ClassVar

from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings


class BackendApiStatus(Enum):
    OK = auto()
    INVALID_COMMAND = auto()
    NOT_IMPLEMENTED = auto()
    COMMAND_FAILED = auto()
    TIMEOUT = auto()


WallIndex = int
FlatGridIndex = int
FanLayerIndex = int
size_2d = tuple[int, int]
AcceptablePWMData = dict[FanLayerIndex, list[int]]
AcceptableRPMData = dict[FanLayerIndex, list[int]]


class ModuleType(str, Enum):
    """
    Enum representing the type of a Windshaper Module.
    """

    MODULE_0812 = "0812"
    MODULE_0816 = "0816"
    MODULE_0818 = "0818"
    MODULE_2420 = "2420"


class ModuleTypeConfig(BaseSettings):
    type: str = Field(..., alias="type")
    number_of_layers: int = Field(..., alias="number_of_layers")
    number_fans_per_layer: int = Field(..., alias="number_fans_per_layer")
    layers_rotation_directions: dict[str, str] = Field(..., alias="layers_rotation_directions")
    max_power_watt: int = Field(..., alias="max_power_watt")


class ControllableModule(BaseModel):
    MAX_LIFEPOINTS: ClassVar[int] = 50

    mac: str
    ip: str | None = None
    type: ModuleType
    # config: ModuleTypeConfig | None

    # ! States
    rpms: AcceptableRPMData
    is_connected: bool = False
    is_blinking: bool = False
    is_psu: bool = False
    lifepoints: int = MAX_LIFEPOINTS

    # ! Commands
    pwms: AcceptablePWMData
    blink_command: bool = False
    psu_command: bool = False


class WalledModules(BaseModel):
    """
    Represents walls of modules

    Each wall contains a dictionary of modules indexed in the wall (0, 0 is top-left corner)
    """

    walls: dict[WallIndex, dict[FlatGridIndex, ControllableModule]]


class WalledModulesBuilder:
    def __init__(
        self,
        walled_modules: WalledModules,
        wall_sizes: dict[WallIndex, size_2d],
        module_sizes: dict[tuple[WallIndex, FlatGridIndex], size_2d],
        module_nb_layers: dict[tuple[WallIndex, FlatGridIndex], int],
    ):
        self._walled_modules: WalledModules = walled_modules
        self.wall_sizes: dict[WallIndex, size_2d] = wall_sizes
        self.module_sizes: dict[tuple[WallIndex, FlatGridIndex], size_2d] = module_sizes
        self.module_nb_layers: dict[tuple[WallIndex, FlatGridIndex], int] = module_nb_layers

    def column(self, col: int) -> WalledModulesBuilder:
        # Select a single column for all (or selected) walls
        return self.cols([col])

    def module_flat(self, flat_index: int) -> WalledModulesBuilder._ModuleContext:
        # Only valid if a single wall is selected
        wall_idxs = list(self.wall_sizes.keys()) if len(self.wall_sizes) == 1 else None
        if wall_idxs is None:
            raise ValueError("module_flat() requires a single wall to be selected")
        wall = self._walled_modules.walls[wall_idxs[0]]
        if flat_index not in wall:
            raise ValueError(f"flat_index {flat_index} not in wall")
        return WalledModulesBuilder._ModuleContext(wall[flat_index])

    def module(self, x: int, y: int, ncols: int | None = None) -> WalledModulesBuilder._ModuleContext:
        # Only valid if a single wall is selected
        wall_idxs = list(self.wall_sizes.keys()) if len(self.wall_sizes) == 1 else None
        if wall_idxs is None:
            raise ValueError("module() requires a single wall to be selected")
        wall = self._walled_modules.walls[wall_idxs[0]]
        if ncols is None:
            ncols = self.wall_sizes[wall_idxs[0]][0] if wall_idxs[0] in self.wall_sizes else self._infer_ncols(wall)
        flat_index = y * ncols + x
        if flat_index not in wall:
            raise ValueError(f"flat_index {flat_index} (from x={x}, y={y}, ncols={ncols}) not in wall")
        return WalledModulesBuilder._ModuleContext(wall[flat_index])

    # Add column and module to wall context for chaining
    def _wall_context_column(self, col: int) -> WalledModulesBuilder:
        # Only valid if a single wall is selected
        if len(self.wall_sizes) != 1:
            raise ValueError("column() requires a single wall to be selected")
        return self.cols([col])

    def _wall_context_module(
        self, flat_index: int | None = None, x: int | None = None, y: int | None = None, ncols: int | None = None
    ) -> WalledModulesBuilder._ModuleContext:
        if flat_index is not None:
            return self.module_flat(flat_index)
        if x is not None and y is not None:
            return self.module(x, y, ncols)
        raise ValueError("Either flat_index or both x and y must be provided")

    def wall(self, wall_index: WallIndex) -> WalledModulesBuilder:
        # Create new wall_sizes, module_sizes, module_nb_layers for the selected wall
        wall_sizes = {wall_index: self.wall_sizes[wall_index]} if wall_index in self.wall_sizes else {}
        module_sizes = {k: v for k, v in self.module_sizes.items() if k[0] == wall_index}
        module_nb_layers = {k: v for k, v in self.module_nb_layers.items() if k[0] == wall_index}
        return WalledModulesBuilder(
            self._walled_modules,
            wall_sizes=wall_sizes,
            module_sizes=module_sizes,
            module_nb_layers=module_nb_layers,
        )

    def walls(self, wall_indexs: list[WallIndex]) -> WalledModulesBuilder:
        wall_sizes = {k: v for k, v in self.wall_sizes.items() if k in wall_indexs}
        module_sizes = {k: v for k, v in self.module_sizes.items() if k[0] in wall_indexs}
        module_nb_layers = {k: v for k, v in self.module_nb_layers.items() if k[0] in wall_indexs}
        return WalledModulesBuilder(
            self._walled_modules,
            wall_sizes=wall_sizes,
            module_sizes=module_sizes,
            module_nb_layers=module_nb_layers,
        )

    def rows(self, row: list[int]) -> WalledModulesBuilder:
        # Select all modules in the specified row(s) for each selected wall
        selected_walls = self.wall_sizes.keys()
        new_wall_sizes = {}
        new_module_sizes = {}
        new_module_nb_layers = {}
        new_walls = {}
        for wall_idx in selected_walls:
            ncols, nrows = self.wall_sizes[wall_idx]
            wall = self._walled_modules.walls[wall_idx]
            selected_modules = {}
            for r in row:
                if r < 0 or r >= nrows:
                    continue
                for c in range(ncols):
                    flat_idx = r * ncols + c
                    if flat_idx in wall:
                        selected_modules[flat_idx] = wall[flat_idx]
                        k = (wall_idx, flat_idx)
                        if k in self.module_sizes:
                            new_module_sizes[k] = self.module_sizes[k]
                        if k in self.module_nb_layers:
                            new_module_nb_layers[k] = self.module_nb_layers[k]
            if selected_modules:
                new_walls[wall_idx] = selected_modules
                new_wall_sizes[wall_idx] = self.wall_sizes[wall_idx]
        new_walled_modules = WalledModules(walls=new_walls)
        return WalledModulesBuilder(
            new_walled_modules,
            wall_sizes=new_wall_sizes,
            module_sizes=new_module_sizes,
            module_nb_layers=new_module_nb_layers,
        )

    def cols(self, cols: list[int]) -> WalledModulesBuilder:
        # Select all modules in the specified column(s) for each selected wall
        selected_walls = self.wall_sizes.keys()
        new_wall_sizes = {}
        new_module_sizes = {}
        new_module_nb_layers = {}
        new_walls = {}
        for wall_idx in selected_walls:
            ncols, nrows = self.wall_sizes[wall_idx]
            wall = self._walled_modules.walls[wall_idx]
            selected_modules = {}
            for c in cols:
                if c < 0 or c >= ncols:
                    continue
                for r in range(nrows):
                    flat_idx = r * ncols + c
                    if flat_idx in wall:
                        selected_modules[flat_idx] = wall[flat_idx]
                        k = (wall_idx, flat_idx)
                        if k in self.module_sizes:
                            new_module_sizes[k] = self.module_sizes[k]
                        if k in self.module_nb_layers:
                            new_module_nb_layers[k] = self.module_nb_layers[k]
            if selected_modules:
                new_walls[wall_idx] = selected_modules
                new_wall_sizes[wall_idx] = self.wall_sizes[wall_idx]
        new_walled_modules = WalledModules(walls=new_walls)
        return WalledModulesBuilder(
            new_walled_modules,
            wall_sizes=new_wall_sizes,
            module_sizes=new_module_sizes,
            module_nb_layers=new_module_nb_layers,
        )

    def set(self, pwm: int) -> WalledModulesBuilder:
        wall_indexs = self._wall_indexs if self._wall_indexs is not None else list(self._walled_modules.walls.keys())
        for wall_index in wall_indexs:
            wall = self._walled_modules.walls[wall_index]
            if self._row_idxs is not None or self._col_idxs is not None:
                # Need to know ncols for flat index math
                ncols = self._infer_ncols(wall)
                nrows = self._infer_nrows(wall, ncols)
                row_range = self._row_idxs if self._row_idxs is not None else list(range(nrows))
                col_range = self._col_idxs if self._col_idxs is not None else list(range(ncols))
                for row in row_range:
                    for col in col_range:
                        flat_idx = row * ncols + col
                        if flat_idx in wall:
                            self._apply_set(wall[flat_idx], pwm)
            else:
                for module in wall.values():
                    self._apply_set(module, pwm)
        return self

    def _infer_ncols(self, wall: dict[int, ControllableModule]) -> int:
        # Infer ncols from the max flat index (assume rectangular grid)
        if not wall:
            return 0
        max_idx = max(wall.keys())
        # Try to guess ncols by finding the smallest difference between indices
        # This assumes a row-major order
        # If user provides ncols, this can be improved
        return int(max_idx**0.5) + 1

    def _infer_nrows(self, wall: dict[int, ControllableModule], ncols: int) -> int:
        if not wall or ncols == 0:
            return 0
        return (max(wall.keys()) // ncols) + 1

    # The rest of the API (module, layer, fan, etc.) can be added as needed, following the same pattern
    # ...existing code for _ModuleContext, _LayerContext, _FanContext, etc...

    class _ModuleContext:
        def __init__(self, module: ControllableModule):
            self._module: ControllableModule = module

        def layer(self, layer_idx: int) -> WalledModulesBuilder._LayerContext:
            return WalledModulesBuilder._LayerContext(self._module, layer_idx)

        def fan_flat(self, flat_index: int) -> WalledModulesBuilder._FanContext:
            return WalledModulesBuilder._FanContext(self._module, flat_index)

        def fan(self, x: int, y: int, ncols: int) -> WalledModulesBuilder._FanContext:
            flat_index: int = y * ncols + x
            return WalledModulesBuilder._FanContext(self._module, flat_index)

        def set(self, pwm: int) -> WalledModulesBuilder._ModuleContext:
            WalledModulesBuilder._apply_set(self._module, pwm)
            return self

    class _LayerContext:
        def __init__(self, module: ControllableModule, layer_idx: int) -> None:
            self._module: ControllableModule = module
            self._layer_idx: int = layer_idx
            self._layers: Sequence[int] = [layer_idx]

        def layers(self, layers: Sequence[int]) -> WalledModulesBuilder._LayerContext:
            self._layers = layers
            return self

        def set(self, pwm: int) -> WalledModulesBuilder._LayerContext:
            for layer in self._layers:
                WalledModulesBuilder._apply_set(self._module, pwm, layer=layer)
            return self

    class _FanContext:
        def __init__(self, module: ControllableModule, flat_index: int) -> None:
            self._module: ControllableModule = module
            self._flat_index = flat_index
            self._layers: Sequence[int] | None = None

        def layers(self, layers: Sequence[int]) -> WalledModulesBuilder._FanContext:
            self._layers = layers
            return self

        def set(self, pwm: int) -> WalledModulesBuilder._FanContext:
            layers = self._layers or getattr(self._module, "pwms", {}).keys()
            for layer in layers:
                WalledModulesBuilder._apply_set(self._module, pwm, layer=layer, fan=self._flat_index)
            return self

    @staticmethod
    def _apply_set(module: ControllableModule, pwm: int, layer: int | None = None, fan: int | None = None) -> None:
        raise NotImplementedError("Implement this method to set PWM at the desired granularity.")




if __name__ == "__main__":
    # ! Example for the builder

    # 1. Set a full wall to PWM 500
    walled_modules = WalledModules(
        walls={
            0: {
                0: ControllableModule(
                    mac="00:11:22:33:44:55",
                    type=ModuleType.MODULE_0812,
                    rpms={0: [0] * 12},
                    pwms={0: [0] * 12},
                ),
                1: ControllableModule(
                    mac="66:77:88:99:AA:BB",
                    type=ModuleType.MODULE_0812,
                    rpms={0: [0] * 12},
                    pwms={0: [0] * 12},
                ),
            },
            1: {
                0: ControllableModule(
                    mac="CC:DD:EE:FF:00:11",
                    type=ModuleType.MODULE_0812,
                    rpms={0: [0] * 12},
                    pwms={0: [0] * 12},
                ),
                1: ControllableModule(
                    mac="22:33:44:55:66:77",
                    type=ModuleType.MODULE_0812,
                    rpms={0: [0] * 12},
                    pwms={0: [0] * 12},
                ),
            },
        }
    )


WallIndex = int
PWMValue = int
FlatGridIndex = int
FanLayerIndex = int
size_2d = tuple[int, int]
AcceptablePWMData = dict[FanLayerIndex, list[int]]
AcceptableRPMData = dict[FanLayerIndex, list[int]]


class WalledModules2(BaseModel):
    """
    Represents walls of modules

    Each wall contains a dictionary of Values indexed in the wall (0, 0 is top-left corner)
    """

    walls: dict[WallIndex, dict[FlatGridIndex, PWMValue]]

    wall_sizes: dict[int, tuple[int, int]] = {0: (2, 1), 1: (2, 1)}  # 2 columns, 1 row for each wall
    module_sizes: dict[tuple[int, int], tuple[int, int]] = {
        (0, 0): (1, 1),
        (0, 1): (1, 1),
        (1, 0): (1, 1),
        (1, 1): (1, 1),
    }
    module_nb_layers: dict[tuple[int, int], int] = {
        (0, 0): 1,
        (0, 1): 1,
        (1, 0): 1,
        (1, 1): 1,
    }

    builder = WalledModulesBuilder(
        walled_modules=walled_modules,
        wall_sizes=wall_sizes,
        module_sizes=module_sizes,
        module_nb_layers=module_nb_layers,
    )

    builder.set(pwm=500) # Set all walls to 500
    builder.wall(wall_index=0).set(pwm=500) # Set wall 0 to 500
    builder.wall(wall_index=1).set(pwm=500) # Set wall
    builder.wall(wall_index=1).set(pwm=500) # Set wall


    builder.wall(wall_index=0).set(pwm=500)
    builder.wall(wall_index=1).cols(col=[0, 2]).set(pwm=500) # set column 0 and 2 of wall 1 to 500
    builder.wall(wall_index=0).rows(row=[0]).set(pwm=500)
    builder.wall(wall_index=0).rows(row=[0, 1]).set(pwm=500)
    builder.wall(wall_index=0).rows(row=[0, 1]).cols(cols=[0]).set(pwm=500)  # set rows 0 and 1 of column 0 to 500
    builder.wall(wall_index=0).rows(row=[0, 1]).cols(cols=[0, 2]).set(pwm=500)  # set rows 0 and 1 of column 0 and column 2 to 500

    builder.rows(row=[0]).cols(cols=[0, 1]).set(pwm=500)  # set row 0 of column 0 and 1 of all walls, to 500
    builder.cols(cols=[0, 1]).set(pwm=500)  # set column 0 and 1 of all walls, to 500
    builder.wall(wall_index=0).cols(cols=[0]).modules([0, 1]).set(pwm=500) # set modules 0 and 1 of column 0 of wall 0 to 500 (module is flat index)
    builder.wall(wall_index=2).cols(cols=[0]).modules([0]).layers(layers=[0, 1]).set(pwm=500)
    builder.wall(wall_index=0).cols(cols=[0]).modules([0]).fans(flat_index=[0]).set(pwm=500)

    builder.wall(wall_index=0).cols(cols=[0]).modules(flat_indexes=[0]).fans(flat_indexes=[0]).set(pwm=500)
    builder.wall(wall_index=0).cols(cols=[0]).modules(flat_indexes=[0]).fans(flat_indexes=[0]).layers(layers=[0, 1]).set(pwm=500)
    builder.wall(wall_index=0).cols(cols=[0]).modules(flat_indexes=[0]).fans(flat_indexes=[0]).layers(layers=[0, 1]).set(pwm=500)

    # 2D index version
    builder.wall(wall_index=0).cols(cols=[0]).rows(rows=[1, 2]).modules_2d(indexes=[(0, 1), (0, 2)]).fans(indexes=[(0, 0), (0, 2)]).set(pwm=500)
    builder.wall(wall_index=0).cols(cols=[0]).modules_2d(indexes=[(0, 1), (0, 2)]).fans(indexes=[(0, 0), (0, 2)]).layers(layers=[0, 1]).set(pwm=500)
    builder.wall(wall_index=0).cols(cols=[0]).modules_2d(indexes=[(0, 1), (0, 2)]).fans(indexes=[(0, 0), (0, 2)]).layers(layers=[0, 1]).set(pwm=500)



from pydantic import BaseModel

from .common_models import Vec3


class WindProbeData(BaseModel):
    """
    Represents data from a wind probe.
    """

    timestamp_s: float
    wind_velocity_mps_probe_ref: Vec3  # Raw data
    wind_velocity_mps_windshaper_ref: Vec3
    wind_velocity_mps_windshaper_ref_corrected: Vec3  # Corrected for probe movement

    temperature_celcius: float
    atmospheric_pressure_hpascal: float
    static_pressure_pascal: float



from typing import NamedTuple


class ModuleUpdate(NamedTuple):
    mac: str
    ip: str
    type: dict[str, str]
    rpms: dict[int, list[int]]
    is_connected: bool
    is_blinking: bool
    is_psu: bool
    lifepoints: int
    pwms: dict[int, list[int]]
    blink_command: bool
    psu_command: bool
